#!/usr/bin/env python

# draws a wireframe of an obj file onto active layer
# Author: Tin Tran 
# Date: 2014

import math
import string
from gimpfu import *
from array import array
def degtorad(angle):
	return (1.0 * angle / 180 * math.pi)
	
def is_number(s):
	try:
		float(s)
		return True
	except ValueError:
		return False

def read_vector(parts):
	vector = []
	for part in parts:
		if (is_number(part)):
			vector.append(float(part))
	return vector

def read_face(parts):
	face = []
	for part in parts:
		if (is_number(part)):
			face.append(int(part))
		else: #check and see if it's multiple values just read first value
			newparts = part.split('/')
			if (is_number(newparts[0])):
				face.append(int(newparts[0]))
			else:    #check and see if it starts with a 'v' which stands for vertex
				vparts = newparts[0].split('v')
				if (is_number(vparts[0])):
					face.append(int(vparts[0]))
	return face
	
def rotate_vectors(v, rotate_x, rotate_y, rotate_z):
	rotated_vectors = []       #return array
	radx = degtorad(rotate_x)
	rady = degtorad(rotate_y)
	radz = degtorad(rotate_z)
	
	for vec in v:
		rvec = [0,0,0]
	    #rotate X axis
		tmp = [vec[0],vec[1],vec[2]]
		rvec[0] = tmp[0]
		rvec[1] = math.cos(radx)*tmp[1] - math.sin(radx)*tmp[2]
		rvec[2] = math.sin(radx)*tmp[1] - math.cos(radx)*tmp[2]
		#rotate Y axis
		tmp = [rvec[0],rvec[1],rvec[2]]
		rvec[0] = math.cos(rady)*tmp[0] + math.sin(rady)*tmp[2]
		rvec[1] = tmp[1]
		rvec[2] = -math.sin(rady)*tmp[0] + math.cos(rady)*tmp[2]
		#rotate Z axis
		tmp = [rvec[0],rvec[1],rvec[2]]
		rvec[0] = math.cos(radz)*tmp[0] - math.sin(radz)*tmp[1]
		rvec[1] = math.sin(radz)*tmp[0] + math.cos(radz)*tmp[1]
		rvec[2] = tmp[2]
		#add it to rotated vectors
		rotated_vectors.append(rvec)
	return (rotated_vectors)
def project_2D(v,midx,midy,BZ,SCALE,autoScale): #returns projected vectors,minZ,maxZ
	minZ =  32000
	maxZ = -32000
	maxX = 0 #used to determine auto-scale
	maxY = 0 #used to determine auto-scale
	indexX = -1 #used to point to which 2D vertice is furthest from midx
	indexY = -1 #used to point to which 2D vertice is furthest from midy
	projected_vectors = []
	for i,vec in enumerate(v):
		tmp = [0,0,0]
		tmp[0] = midx + vec[0]*BZ/(vec[2]+BZ)*SCALE
		tmp[1] = midy + vec[1]*BZ/(vec[2]+BZ)*SCALE
		tmp[2] = vec[2]           #just copy z value for depth decisions later
		projected_vectors.append(tmp)  #add it to projected vectors
		minZ = min(minZ,tmp[2])
		maxZ = max(maxZ,tmp[2])
		#auto scale option ----------------------------------
		currentX = abs(vec[0]*BZ/(vec[2]+BZ)*SCALE)  #see how far it is from midx
		if (currentX > maxX):
			maxX = currentX
			indexX = i
		currentY = abs(vec[1]*BZ/(vec[2]+BZ)*SCALE)
		if (currentY > maxY):
			maxY = currentY
			indexY = i
	if (autoScale == True):
		scaleX = (9.0 * midx/10)/abs(v[indexX][0]*BZ/(v[indexX][2]+BZ))  #calculate scaleX to fit screen 9/10 of screen
		scaleY = (9.0 * midy/10)/abs(v[indexY][1]*BZ/(v[indexY][2]+BZ))  #calculate scaleY to fit screen 9/10 of screen
		pdb.gimp_message ("ScaleX: " + str(scaleX))
		pdb.gimp_message ("ScaleY: " + str(scaleY))
		#Change scale value and re-calculate/re-project
		SCALE = min(scaleX,scaleY) #use the smaller scale out of the two to fit screen
		pdb.gimp_message ("auto-Scale: " + str(SCALE))
		projected_vectors = []
		for i,vec in enumerate(v):
			tmp = [0,0,0]
			tmp[0] = midx + vec[0]*BZ/(vec[2]+BZ)*SCALE
			tmp[1] = midy + vec[1]*BZ/(vec[2]+BZ)*SCALE
			tmp[2] = vec[2]                  #just copy z value for depth decisions later
			projected_vectors.append(tmp)    #add it to projected vectors
	return (projected_vectors,minZ,maxZ)	
def putpixel(src_pixels,x,y,image_width,p_size,r,g,b):
	src_pos = (x + image_width * y) * p_size
	pixel = src_pixels[src_pos: src_pos + p_size]
	pixel[0] = r
	pixel[1] = g
	pixel[2] = b
	if (p_size==4): pixel[3] = 255
	src_pixels[src_pos: src_pos + p_size] = pixel
	return
#line function copied from http://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
#Simplification method
def drawline(src_pixels,x0,y0,x1,y1,image_width,p_size,r,g,b):
	x0 = int(round(x0,0))
	y0 = int(round(y0,0))
	x1 = int(round(x1,0))
	y1 = int(round(y1,0))
	dx = abs(x1-x0)
	dy = abs(y1-y0) 
	if (x0 < x1): sx = 1 
	else: sx = -1
	if (y0 < y1): sy = 1 
	else: sy = -1
	err = dx-dy
	while 1:
		#plot(x0,y0)
		putpixel(src_pixels,x0,y0,image_width,p_size,r,g,b)
		if ((x0 == x1) and (y0 == y1)): break
		e2 = 2*err
		if (e2 > -dy):
			err = err - dy
			x0 = x0 + sx
		if (e2 < dx):
			err = err + dx
			y0 = y0 + sy 
	return

def python_open_obj_file(image, layer, filename, rotatex, rotatey, rotatez, autoScale, scale, bz) :
	vectors = []
	vectors.append([0,0,0]) #add an empty point so that faces can point to right indexes starting at 1
	unique_vectors = []
	unique_vectors.append([0,0,0])
	vector_pointers = []      #use pointers to point to unique_vectors, to remove duplicates of vectors
	vector_pointers.append(0) #first pointer, isn't used just point it zero
	faces = []
	savedLines = []
	savedUniqueLines = []
	#create savedLines array of 256 elements holding empty arrays
	for j in range(0, 256):
		savedLines.append([])
		savedUniqueLines.append([])
	pdb.gimp_image_undo_group_start(image)
	
	image_width = pdb.gimp_image_width(image)
	image_height = pdb.gimp_image_height(image)
	pdb.gimp_message("Width:" + str(layer.width))
	pdb.gimp_message("Height:" + str(layer.height))
	gimp.progress_init("detecting white...")
	gimp.progress_update(1.0 * 100)
	pdb.gimp_progress_end()
	f = open(filename, 'r')
	while 1:
		line=f.readline()
		if not line: break
		parts = line.split(' ')
		if (parts[0] == "v"): #if it's vector read it
			vectors.append(read_vector(parts))
		if (parts[0] == "f"): #if it's a face read it
			faces.append(read_face(parts))		
	f.close()
	'''
	gimp.progress_init("finding unique vectors...")
	vectors_len = len(vectors)
	for i in range(1,len(vectors)): #find unique vectors and add it
		vector_pointers.append(0) #default to zero we'll set it later
		gimp.progress_update(1.0 * i/vectors_len)
		found = 0
		for j in range(1,len(unique_vectors)):
			if (vectors[i][0] == unique_vectors[j][0] and vectors[i][1] == unique_vectors[j][1] and vectors[i][2] == unique_vectors[j][2]):
				if (found == 0):
					found = 1
					vector_pointers[i] = j #make it point to found copy of unique_vectors
		if (found == 0):
			vector_pointers[i] = len(unique_vectors)        #make it point to the current index
			unique_vectors.append(vectors[i])
	pdb.gimp_progress_end()
	
	#debug
	pdb.gimp_message("vectorSize:"+ str(len(vectors)))
	pdb.gimp_message("uniqueVectorSize:"+str(len(unique_vectors)))
	
	#redefine faces to make it use vector_pointers to point to unique_vectors instead of vectors
	for j in range(0,len(faces)):
		for i in range(0,len(faces[j])):
			faces[j][i] = vector_pointers[faces[j][i]]
	vectors = unique_vectors	#set to unique vectors so we operate on unique_vectors from now on.
	'''	
	rotated = rotate_vectors(vectors, rotatex, rotatey, rotatez)
	(projected,minZ,maxZ) = project_2D(rotated,image_width/2,image_height/2,bz,scale,autoScale)
	rangeZ = maxZ - minZ
	
	#make face to end with starting point as well so it's wrapped fully around
	for i,face in enumerate(faces):
		face.append(face[0])
	
	#saves the user's brush to set it back later
	userbrushname = pdb.gimp_context_get_brush()
	#create our own brush
	brushname = pdb.gimp_brush_new("obj pen")
	brushradius = pdb.gimp_brush_set_radius(brushname,.5)
	brushshape = pdb.gimp_brush_set_shape(brushname,0)
	brushhardness = pdb.gimp_brush_set_hardness(brushname,1)
	pdb.gimp_context_set_brush(brushname)
	
	#sort out colorValue based on depth(z) and save it to savedLines (ordered)
	for face in faces:
		for j in range(0, (len(face)-1)):
			colorValue = int(255 - ((projected[face[j]][2] + projected[face[j+1]][2])/2-minZ)/rangeZ * 230)
			savedLines[colorValue].append([face[j],face[j+1]])
			
	#copy Uniquelines so we don't have to draw it multiple times
	'''
	gimp.progress_init("finding unique edges...")
	for k in range(0,256):
		gimp.progress_update(1.0 * k/255)
		for line in savedLines[k]:
			found = 0
			for findline in savedUniqueLines[k]:
				if (((line[0] == findline[0]) and (line[1] == findline[1])) or ((line[0] == findline[1]) and (line[1] == findline[0]))):
					found = 1
			if (found == 0):  #not found, we add it to unique line list
				savedUniqueLines[k].append(line)
	pdb.gimp_progress_end()		

	lineCount = 0
	for k in range(0,256):
		lineCount = lineCount + len(savedLines[k])
	pdb.gimp_message("sizeOfSavedLines:" + str(lineCount))
	
	lineCount2 = 0
	for k in range(0,256):
		lineCount2 = lineCount2 + len(savedUniqueLines[k])
	pdb.gimp_message("sizeOfSavedUniqueLines:" + str(lineCount2))
	'''
	#draw it from darkest to brightest (previously calculated above)
	gimp.progress_init("drawing...")
	srcRgn = layer.get_pixel_rgn(0, 0, image_width, image_height,True, False)
	src_pixels = array("B", srcRgn[0:image_width, 0:image_height])
	p_size = len(srcRgn[0,0])
	for j in range(0,256):
		gimp.progress_update(1.0 * j/255)
		#pdb.gimp_context_set_foreground((j,j,j))
		for line in savedLines[j]:
			drawline(src_pixels,projected[line[0]][0],projected[line[0]][1],projected[line[1]][0],projected[line[1]][1],image_width,p_size,j,j,j)
			#pdb.gimp_pencil(layer,4,[projected[line[0]][0],projected[line[0]][1],projected[line[1]][0],projected[line[1]][1]])
	pdb.gimp_progress_end()
	
	
	#test put pixel
	#putpixel(src_pixels,10,10,image_width,p_size,255,255,255)
	#putpixel(src_pixels,11,10,image_width,p_size,255,0,0)
	#putpixel(src_pixels,12,10,image_width,p_size,0,255,0)
	#putpixel(src_pixels,13,10,image_width,p_size,0,0,255)
	
	#copy it back to srcRgn
	srcRgn[0:image_width, 0:image_height] = src_pixels.tostring()
	
	
	layer.flush()
	#layer.merge_shadow(True)
	layer.update(0, 0, image_width,image_height)
	pdb.gimp_message(p_size)
	
	
	#delete brush when done
	pdb.gimp_brush_delete(brushname)
	pdb.gimp_context_set_brush(userbrushname) #set it back so we don't surprise user
		
	pdb.gimp_message("Completed!")
	pdb.gimp_image_undo_group_end(image)
	pdb.gimp_displays_flush()
    #return

register(
    "python_fu_open_obj_file",
    "Open an obj file",
    "Open obj file and draws it on active layer",
    "Tin Tran",
    "Tin Tran",
    "2014",
    "<Image>/File/Open Object file...",
    "RGB*, GRAY*",      # Create a new image, don't work on an existing one
    [
	(PF_FILE, "infilename", "File to open", "/Default/Path"),
	(PF_ADJUSTMENT, "rotatex", "Rotate X-axis", 270 ,(0,360,1)),
	(PF_ADJUSTMENT, "rotatey", "Rotate Y-axis", 0 ,(0,360,1)),
	(PF_ADJUSTMENT, "rotatez", "Rotate Z-axis", 0 ,(0,360,1)),
	(PF_TOGGLE, "auto_scale",   ("Auto-scale to fit image?"),   True),
	(PF_ADJUSTMENT, "scale", "Scale", 160 ,(1,5000,.1)),
	(PF_ADJUSTMENT, "bz", "BZ distance", 600 ,(1,1000,10))
	],
    [],
    python_open_obj_file)

main()
